<?php
/*******************************************************************************
 *	EmailQueue.php	Email queue module
 *			T.Barnaby,	BEAM Ltd,	2016-01-26
 *******************************************************************************
 *
 * This accepts emails into a queue in the database.
 * A cron job will call this module to send about 75 emails at a time. This
 * is to stop SPAM errors happending and to save PHP timeouts occuring.
 *
 * If an email is sent to more than one email address, then a single copy of
 * the headers and message are stored in the first email record, and this is then
 * used when sending all of the other emails. This is to save database space and
 * time when creating the emails to send.
 *
 * Testing
 *	Setting the objects test variable disables the actual sending of the emails.
 *
 */
require_once "BModule.php";
require_once "BBlock.php";
require_once "BView.php";
require_once "BDate.php";
require_once "BEdit.php";
require_once "BDate.php";

class BObjEmailQueue extends BObj {
	function __construct(){
		parent::__construct(array(
			new BMember("id", "Int", "Int"),
			new BMember("dateAdded", "DateTime", "Datetime"),
			new BMember("emailAddress", "String"),
			new BMember("masterId", "Int"),
			new BMember("headers", "StringHtml"),
			new BMember("subject", "String"),
			new BMember("message", "StringHtml", Null, Null, "longtext"),
			new BMember("sent", "Int"),
			new BMember("error", "Int"),
			new BMember("userName", "String"),
		));
	}
};

class BObjEmailForward extends BObj {
	function __construct(){
		parent::__construct(array(
			new BMember("id", "Int", "Int"),
			new BMember("dateAdded", "DateTime", "Datetime"),
			new BMember("emailAddress", "String"),
			new BMember("forwards", "String"),
		));
	}
};

class EmailQueue extends BModule {
	// Basic Interface routines
	function __construct(&$sys){
		parent::__construct($sys);
		$this->name = "EmailQueue";
		$this->title = "EmailQueue";
		$this->info = "EmailQueue module. This module queues emails and sends them over a period of time from a cron job.";
		$this->test = 0;
		
		// Add Objects
		$this->sys->db->addObjectClasses(array("EmailQueue" => "BObjEmailQueue"));
		$this->sys->db->addObjectClasses(array("EmailForwards" => "BObjEmailForward"));
	}
	
	function checkAccess(){
		return ($this->sys->login->checkLogin("admin"));
	}
	
	function getMenu(){
		if($this->checkAccess()){
			return new BMenu("EmailQueue", "/{$this->name}", "", array("EmailQueue List" => "/{$this->name}"));
		}
		else {
			return Null;
		}
	}

	function getPage($uri){
		$this->parseUri($uri);

		$page = "";

		if($this->checkAccess()){
			switch($this->cmd){
			case "view":			$page = $this->view();	break;
			case "sendQueuedEmails":	$page = $this->sendQueuedEmails();	break;
			case "clean":			$page = $this->clean();	break;
			case "cleanAllSent":		$page = $this->cleanAllSent();	break;
			case "forwardsEdit":		$page = $this->forwardsEdit();	break;
			case "forwardsAdd":		$page = $this->forwardsAdd();	break;
			case "forwardsDelete":		$page = $this->forwardsDelete();	break;
			case "forwardsSave":		$page = $this->forwardsSave();	break;
			default: 			$page = $this->viewList(); break;
			}
			return $page;
		}
		else {
			switch($this->cmd){
			case "sendQueuedEmails":	$page = $this->sendQueuedEmails();	break;
			default:
		 		$this->sys->pageDisplayError(new BError(403, "You are not allowed to access EmailQueue"));
				break;
			}
			return $page;
		}
	}

	function getPageBlock($uri){
		$o = Null;
		if($this->checkAccess()){
			$help = $this->sys->baseUrl . "/modules/EmailQueue/info.html";

			$o = new BMenu("EmailQueue", "", "", array(
				new BMenu("List", "", 			"href='?cmd=viewList'"),
				new BMenu("Send Queued Emails", "",	"href='?cmd=sendQueuedEmails'"),
				new BMenu("Email forwards edit", "", 	"href='?cmd=forwardsEdit'"),
				new BMenu("------", "", ""),
				new BMenu("Clean old entries", "", 	"href='?cmd=clean'"),
				new BMenu("Clean All Sent", "", 	"href='?cmd=cleanAllSent'"),
				new BMenu("------", "", ""),
				new BMenu("Help", "",			"href='{$help}'"),
				));
		}
		return $o;
	}

	function viewList(){
		$page = "<h2>EmailQueue information</h2>\n";
		
		if(!$this->checkAccess()){
			$this->sys->pageDisplayError(new BError(1, "You are not allowed to access EmailQueue"));
		}
		
//		print "Form: "; print_r($_POST); print "<br>\n";
		
		$sel = new BSelect("select", "Select", "", 3);
		$sel->setView(array(
			new BViewTypeSelect("period", "dateAdded", "Period", "period", "select", array("", "Last Year", "Last 6 Months", "Last 2 Months", "Last Month", "Last Week")),
			new BViewTypeSelect("startDate", "dateAdded", "Startdate", "greaterThan", "date"),
			new BViewTypeSelect("endDate", "dateAdded", "EndDate", "lessThanEqual", "date"),
			new BViewTypeSelect("sent", "sent", "Sent", "equal", "tristate"),
			new BViewTypeSelect("userName", "userName", "From User", "like"),
			new BViewTypeSelect("emailAddress", "emailAddress", "To Email address", "like"),
			new BViewTypeSelect("masterId", "masterId", "Master Only", "isZero", "checkbox"),
			new BViewTypeSelect("pager", "", "Page", "pager", "pager"),
			));

		$sel->getObj()->period = "Last Month";
		$sel->getObj()->masterId = 1;
		$sel->pagerSet(0, 50, 0);
		$sel->postUpdate();
		
		$where = $sel->getSelectString();

//		print "Where: {$where}<br>\n";
//		BDatabase::$debug = 1;
		
		$view = new BViewList("list", "action");
		$view->setView(array(
			new BViewType("id"),
			new BViewType("dateAdded", "Date Added", "datetime"),
			new BViewType("masterId", "MasterId"),
			new BViewType("emailAddress", "EmailAddress"),
			new BViewType("subject", "Subject"),
			new BViewType("sent", "Sent"),
			new BViewType("error", "Error"),
			new BViewType("userName", "UserName"),
			));
			
		$list = $this->sys->db->getObjectList("EmailQueue", "", $where, "!dateAdded", $sel->pagerGetLimitString());
		$totalNumber = $this->sys->db->getObjectNumber("EmailQueue", $where);
		$sel->pagerSet(0, 50, $totalNumber);

		$view->setList($list, 1);

		$page = $page . $sel->getHtml();
		$page = $page . $view->getHtml();

		return $page;
	}
	
	function view(){
		if(!isset($_GET["id"])){
			$this->sys->pageDisplayError(new BError(1, "No object selected"));
			return "";
		}
		else {
			$obj = $this->sys->db->getObject("EmailQueue", filterToInt($_GET["id"]));
		}

		if($obj->masterId){
			$mobj = $this->sys->db->getObject("EmailQueue", $obj->masterId);
			$obj->message = $mobj->message;
		}

		$page = "<h2>Email View</h2>";

		$view = new BViewObj("email");
		$view->setView(array(
			new BViewType("id"),
			new BViewType("dateAdded", "Date Added", "datetime"),
			new BViewType("masterId", "MasterId"),
			new BViewType("emailAddress", "EmailAddress"),
			new BViewType("subject", "Subject"),
			new BViewType("sent", "Sent"),
			new BViewType("error", "Error"),
			new BViewType("headers", "", "textarea", "", 400, 200),
			new BViewType("message", "", "textarea", "", 400, 200),
			new BViewType("userName", ""),
			));
		$view->setObj($obj);
		$page = $page . $view->getHtml();

		return $page;
	}


	function sendEmails($emailAddresses, $headers, $subject, $message){
		if(is_string($emailAddresses))
			$emailAddressList = explode(",", $emailAddresses);
		else
			$emailAddressList = $emailAddresses;

		$num = 0;
		$masterId = 0;

		foreach($emailAddressList as $e){
			$obj = $this->sys->db->getNewObject("EmailQueue");
			$obj->dateAdded = strftime("%Y-%m-%d %H:%M:%S");
			$obj->emailAddress = $e;
			$obj->subject = $subject;
			$obj->sent = 0;
			$obj->error = 0;
			if($this->sys->user)
				$obj->userName = $this->sys->user->userName;

			if($num == 0){
				$obj->masterId = 0;
				$obj->headers = $headers;
				$obj->message = $message;
				$this->sys->db->appendObject("EmailQueue", $obj);
				$masterId = $obj->id;
			}
			else {
				$obj->masterId = $masterId;
				$this->sys->db->appendObject("EmailQueue", $obj);
			}
			$num = $num + 1;
		}
		
		// Send the first batch
		$this->sendQueuedEmails();
		
		return 1;
	}
	
	function sendQueuedEmails(){
		$page = "Send queued emails<br/>";
		$mobj = Null;
		
		// Send the next 75 or less emails. Normally called from a cron job
		$list = $this->sys->db->getObjectList("EmailQueue", "", "`sent` = '0' and `error` = '0'", "dateAdded", 75);
		foreach($list as $obj){
			if($obj->masterId){
				if(($mobj == Null) || ($mobj->id != $obj->masterId))
					$mobj = $this->sys->db->getObject("EmailQueue", $obj->masterId);

				if($this->test){
					$page .= "Sendto: {$obj->id} {$obj->emailAddress} {$obj->subject}</br>";
				}
				else {
					$page .= "Sendto: {$obj->id} {$obj->emailAddress} {$obj->subject}</br>";
					$r = mail($obj->emailAddress, $obj->subject, $mobj->message, $mobj->headers);
					if(!$r){
						$page .= "Error sending to: {$obj->emailAddress}<br/>";
						$obj->error = 1;
					}
				}
			}
			else {
				if($this->test){
					$page .= "Sendto: {$obj->id} {$obj->emailAddress} {$obj->subject}</br>";
				}
				else {
					$page .= "Sendto: {$obj->id} {$obj->emailAddress} {$obj->subject}</br>";
//					mail($obj->emailAddress, "Test2: ". $obj->subject, "Hi There", $obj->headers);
					$r = mail($obj->emailAddress, $obj->subject, $obj->message, $obj->headers);
					if(!$r){
						$page .= "Error sending to: {$obj->emailAddress}<br/>";
						$obj->error = 1;
					}
				}
			}
			$obj->sent = 1;
			$this->sys->db->putObject("EmailQueue", $obj);
		}
		
		// Clean up old sent emails
		$this->clean();

		return $page;
	}
	
	function clean(){
		// Delete email records older than 1 month
		$tnow = new BDate("now");
		$tnow->addDays(-30);
		$tnowStr = $tnow->get();

		$list = array(0);
		while($list && count($list)){
			$list = $this->sys->db->getObjectList("EmailQueue", "", "`dateAdded` < '{$tnowStr}'", "", 100);
			foreach($list as $obj){
				$this->sys->db->deleteObject("EmailQueue", $obj->id);
			}
		}
		return "Cleaned emails";
	}
	
	function cleanAllSent(){
		// Delete all sent emails
		$list = array(0);
		while($list && count($list)){
			$list = $this->sys->db->getObjectList("EmailQueue", "", "`sent` = '1'", "", 100);
			foreach($list as $obj){
				$this->sys->db->deleteObject("EmailQueue", $obj->id);
			}
		}
		return "Cleaned emails";
	}
	
	function forwardsEdit(){
		$page = "<h2>Email forwards</h2>\n";

		$view = new BEditDatabase("forwards", "action");

		$where = "";
		
		$view->setView(array(
			new BViewType("emailAddress", ""),
			new BViewType("forwards", "", "text", ["width" => "400"]),
			));

		$view->setSelection($where);
		//$view->setLimit(20);
	
		if($view->order == "")
			$view->setOrder("emailAddress");

		$view->setTable("EmailForwards");

		$page .= "<div class='help'>";
		$page .= "This lists all of the email address forwards of the system. ";
		$page .= "The EmailAddress field should be set to the local email address to forward ie. without the @ and without the domain. ";
		$page .= "The Forwards field should be a comma separated list of email addresses. ";
		$page .= "</div>";

		$page = $page . $view->getHtml();

		return $page;
	}

	function forwardsAdd(){
		$obj = $this->sys->db->getNewObject("EmailForwards");
		$obj->dateAdded = strftime("%Y-%m-%d");
		$this->sys->db->appendObject("EmailForwards", $obj);

		return $this->forwardsEdit();
	}
	function forwardsDelete(){
		if(!isset($_GET["id"])){
			$this->sys->pageDisplayError(new BError(1, "No object selected"));
			return;
		}
		$id = filterToInt($_GET["id"]);
		$obj = $this->sys->db->getObject("EmailForwards", $id);

		// Check role
		if(!$this->checkAccess()){
			$this->sys->pageDisplayError(new BError(1, "You are not authorised to delete this item."));
			exit();
		}

		if(isset($_GET["doIt"]) && $_GET["doIt"] == "1"){
			$this->sys->db->deleteObject("EmailForwards", $id);

			$page = "<h3>{$obj->id}: {$obj->emailAddress} has been Deleted</h3>";
			$page .= $this->forwardsEdit();;
		}
		else {
			$m = new BHtmlNode("div", "name='dialogQuestion'");
			$f = &$m->appendNode("form", "name='dialogQuestion' action='?cmd=forwardsDelete&id={$id}&doIt=1' method='post'");
			$f->appendNode("h3", "", "Do you really want to delete {$obj->emailAddress} ?");

			$f->appendNode("input", "type='submit' name='arg' value='Delete'");
			$f->appendNode("button", "type='button' name='arg' onclick='location.href = \"javascript:history.go(-1)\";'", "Cancel");

			$page = $f->render();
		}

		return $page;
	}

	function forwardsSave(){
		//print "GET: "; print_r($_GET); print "<br>";
		//print "POST: "; print_r($_POST); print "<br>";

		if(isset($_POST["objects"])){
			foreach($_POST["objects"] as $object){
				$obj = new BObjEmailForward();
				$obj->set($object);
				
				//print "Object: "; print_r($obj->getValues()); print "<br>";

				if($err = $obj->validate()){
					$items = implode(",", array_keys($err));
					$this->sys->status = "<span class='editError'>Error in form data in fields: $items</span>";
					return $this->forwardsEdit();
				}
				if(preg_match('/[^a-z_\-.0-9]/i', $obj->emailAddress)){
					$this->sys->status = "<span class='editError'>Error in form data in field: {$obj->emailAddress}</span>";
					return $this->forwardsEdit();
				}
				$emails = explode(",", $obj->forwards);
				foreach($emails as $email){
					if(!filter_var($email, FILTER_VALIDATE_EMAIL)){
						$this->sys->status = "<span class='editError'>Error in form data in field (not a list of email addresses): '{$obj->forwards}'</span>";
						return $this->forwardsEdit();
					}
				}
				
				if($obj->id){
					$this->sys->db->putObject("EmailForwards", $obj);
				}
				else {
					$this->sys->db->appendObject("EmailForwards", $obj);
				}
			}
			
			$page = "<h3>Forwards have been saved into the database</h3>";

			// Save into /etc/aliases
			if($f = fopen("/etc/aliases", "r+")){
				while($line = fgets($f)){
					if($line == "# WebSys forwards\n")
						break;
				}
				
				ftruncate($f, ftell($f));
				
				if(! ($line == "# WebSys forwards\n"))
					fwrite($f, "# WebSys forwards\n");
				
				$list = $this->sys->db->getObjectList("EmailForwards");

				foreach($list as $obj){
					fwrite($f, "{$obj->emailAddress}: {$obj->forwards}\n");
				}
				fclose($f);
				
				exec("/usr/bin/newaliases");

				$page .= "<h3>Saved in /etc/aliases</h3>";
			}
			else {
				$page .= "<h3>Unable to write to /etc/aliases</h3>";
			}
			
		}
		else {
			$page = "<h3>Nothing to save</h3>";
		}
		return $page . $this->forwardsEdit();
	}
};

?>
